@@ -54,7 +54,14 @@ static SCM sym_object = SCM_UNSPECIFIED;
 static SCM sym_args = SCM_UNSPECIFIED;
 static scm_t_bits dynproc_smob_tag = 0;
 
-#if 0 // not useed ATM
+static void gw_guile_handle_wrapper_error(GWLangArena arena,
+                                          GWError *error,
+                                          const char *func_name,
+                                          unsigned int arg_pos) GW_NORETURN;
+static void gw_guile_raise_error (GWLangArena arena, const char *proc,
+                                  const char *error) GW_NORETURN;
+
+#if 0 // not used ATM
 void
 gw_guile_runtime_get_version_info(int *major, int *revision, int *age)
 {
@@ -232,32 +239,36 @@ gw_guile_make_latent_variable (SCM sym, SCM proc, SCM arg)
 
 /* Grr. Seems subrs can't be methods. */
 static void 
-gw_guile_add_subr_method (SCM generic, SCM subr, SCM class_name, SCM module,
+gw_guile_add_subr_method (SCM generic, SCM subr, SCM req_specializers,
                           int n_req_args, int use_optional_args)
 {
   int i;
   char buffer[32];
   SCM specializers, formals, procm, meth, rest_sym = SCM_BOOL_F;
   
+  specializers = SCM_EOL;
+  while (SCM_CONSP (req_specializers))
+  {
+    SCM class_name = SCM_CAR (req_specializers);
+    specializers = scm_cons (SCM_NFALSEP (class_name)
+                             ? SCM_VARIABLE_REF ( scm_lookup (class_name))
+                             : scm_class_top,
+                             specializers);
+    req_specializers = SCM_CDR (req_specializers);
+  }
+  specializers = scm_reverse (specializers);
+  
   if (use_optional_args)
   {
-    formals = SCM_EOL;
     rest_sym = scm_str2symbol ("rest");
-    specializers = scm_class_top;
+    specializers = scm_append_x (scm_list_2 (specializers, scm_class_top));
   }
-  else
-    formals = specializers = SCM_EOL;
   
+  formals = SCM_EOL;
   for (i = n_req_args; i > 0; i--)
   {
     sprintf (buffer, "arg%d", i);
     formals = scm_cons (scm_str2symbol (buffer), formals);
-    if (i == 1)
-      specializers =
-        scm_cons (SCM_VARIABLE_REF (scm_module_lookup (module, class_name)),
-                  specializers);
-    else
-      specializers = scm_cons (scm_class_top, specializers);
   }
 
   if (use_optional_args)
@@ -318,7 +329,7 @@ gw_scm_module_binder_proc (SCM module, SCM sym, SCM definep)
     velts = SCM_VELTS (entry);
 
     gw_guile_add_subr_method (generic,
-                              velts[0], velts[1], velts[2],
+                              velts[0], velts[1],
                               SCM_INUM (velts[3]), SCM_NFALSEP (velts[4]));
 
     proc_list = SCM_CDR (proc_list);
@@ -357,7 +368,7 @@ ensure_scm_module_hacked (void)
 
 /* no explicit returns in this function */
 void
-gw_guile_procedure_to_method_public (SCM proc, SCM class_name,
+gw_guile_procedure_to_method_public (SCM proc, SCM req_specializers,
                                      SCM generic_name,
                                      SCM n_req_args, SCM use_optional_args)
 #define FUNC_NAME "%gw:procedure-to-method-public!"
@@ -366,7 +377,7 @@ gw_guile_procedure_to_method_public (SCM proc, SCM class_name,
   SCM existing_latents;
 
   SCM_VALIDATE_PROC (1, proc);
-  SCM_VALIDATE_SYMBOL (2, class_name);
+  SCM_VALIDATE_LIST (2, req_specializers);
   SCM_VALIDATE_SYMBOL (3, generic_name);
   SCM_VALIDATE_INUM (4, n_req_args);
   /* the fifth is a bool */
@@ -392,10 +403,10 @@ gw_guile_procedure_to_method_public (SCM proc, SCM class_name,
        Otherwise, we're making the first latent binding, and there's nothing in
        the root module that will conflict with our binding. */
     SCM entry = scm_c_make_vector (5, SCM_BOOL_F);
-    /* entry := #(proc class_name module n_req_args use_optional_args) */
+    /* entry := #(proc specializers module n_req_args use_optional_args) */
     
     SCM_VECTOR_SET (entry, 0, proc);
-    SCM_VECTOR_SET (entry, 1, class_name);
+    SCM_VECTOR_SET (entry, 1, req_specializers);
     SCM_VECTOR_SET (entry, 2, scm_current_module ());
     SCM_VECTOR_SET (entry, 3, n_req_args);
     SCM_VECTOR_SET (entry, 4, use_optional_args);
@@ -423,8 +434,8 @@ gw_guile_procedure_to_method_public (SCM proc, SCM class_name,
     }
     else
     {
-      /* The existing binding is something else. We make a new generic, possibly
-         with a different name, and export it. */
+      /* The existing binding is something else. We make a new
+         generic, possibly with a different name, and export it. */
       /* NB: generics also satisfy procedure?. */
       if (SCM_NFALSEP (scm_procedure_p (val)))
       {
@@ -451,7 +462,7 @@ gw_guile_procedure_to_method_public (SCM proc, SCM class_name,
       scm_call_3 (module_add_x, the_root_module, generic_name,
                   scm_make_variable (generic));
     }
-    gw_guile_add_subr_method (generic, proc, class_name, scm_current_module(),
+    gw_guile_add_subr_method (generic, proc, req_specializers, 
                               SCM_INUM (n_req_args),
                               SCM_NFALSEP (use_optional_args));
   }
@@ -479,9 +490,13 @@ dynproc_smob_apply (SCM smob, SCM args)
   error.status = GW_ERR_NONE;
 
   offset = fi->n_req_args * sizeof (void *);
+  rvalue = (void *) ((unsigned char *) data + offset);
+  offset += (fi->ret_type->type->size > sizeof(ffi_arg)
+             ? fi->ret_type->type->size : sizeof(ffi_arg));
   for (i = 0; i < fi->n_req_args; i++)
   {
     SCM arg;
+    offset = GW_ALIGN (offset, fi->arg_types[i]->type->alignment);
     values[i] = (void *) ((unsigned char *) data + offset);
     if (!SCM_CONSP (args))
       scm_wrong_num_args (smob);
@@ -493,10 +508,10 @@ dynproc_smob_apply (SCM smob, SCM args)
     offset += fi->arg_types[i]->type->size;
     args = SCM_CDR (args);
   }
-  rvalue = (void *) ((unsigned char *) data + offset);
   
   ffi_call (&fi->cif, fi->proc, rvalue, values);
 
+  rvalue = GW_RVALUE_PTR (rvalue, fi->ret_type);
   fi->ret_type->wrap_value (&result, ARENA, &fi->ret_typespec, rvalue, &error);
   if (error.status != GW_ERR_NONE)
     gw_handle_wrapper_error (ARENA, &error, fi->proc_name, 0);
@@ -634,11 +649,23 @@ gw_guile_register_wrapset (GWWrapSet *ws)
                                  use_extra_args, (SCM (*)())fi->proc);
     }
     
-    if (fi->generic_name && fi->arg_types && fi->arg_types[0]->class_name)
+    if (fi->generic_name && fi->arg_types)
     {
+      int j;
+      SCM specializers = SCM_EOL;
+      
+      for (j = fi->n_req_args - 1; j >= 0; j--)
+      {
+        GWTypeInfo *arg_type = fi->arg_types[j];
+        SCM class_name = ((arg_type->class_name &&
+                           !(fi->arg_typespecs[j] & GW_TYPESPEC_UNSPECIALIZED))
+                          ? scm_str2symbol(arg_type->class_name) 
+                          : SCM_BOOL_F);
+        specializers = scm_cons (class_name, specializers);
+      }
+      
       gw_guile_procedure_to_method_public (
-              subr,
-              scm_str2symbol (fi->arg_types[0]->class_name),
+              subr, specializers,
               scm_str2symbol (fi->generic_name),
               SCM_MAKINUM (fi->n_req_args),
               (fi->n_optional_args ? SCM_BOOL_T : SCM_BOOL_F));
