@@ -323,7 +323,7 @@
      
      (map
       (lambda (number)
-        (list "  SCM " (out-param-name number) ";\n"))
+        (list "  SCM " (out-param-name number) " = SCM_UNSPECIFIED;\n"))
       (iota (length out-params)))
      
      "\n"
@@ -437,7 +437,10 @@
       (if (null? out-params)
           "  return gw__scm_result;\n"
           (list
-           "  return scm_values (scm_list_n (gw__scm_result, "
+           "  return scm_values (scm_list_n ("
+           (if (needs-result-var? return-type)
+               "gw__scm_result, "
+               '())
            (map (lambda (n)
                   (string-append (out-param-name n) ", "))
                 (iota (length out-params)))
@@ -797,9 +800,18 @@
                 (write
                  `(%gw:procedure->method-public
                    ,(name func)
-                   ',(class-name (first (argument-types func)))
+                   ;; Specializers
+                   ',(map (lambda (arg)
+                            (let ((typespec (typespec arg)))
+                              (and (not (memq 'unspecialized
+                                              (options typespec)))
+                                   (class-name (type typespec)))))
+                          (filter visible? (arguments func)))
                    ',gf
-                   ,(- (argument-count func) (optional-argument-count func))
+                   ;; Required argument count
+                   ,(- (input-argument-count func)
+                       (optional-argument-count func))
+                   ;; Optional arguments?
                    ,(not (zero? (optional-argument-count func))))
                  port)
                 (newline port))
